@@ -34,15 +34,13 @@
 #include "nm-device.h"
 #include "nm-device-wifi.h"
 #include "nm-device-ethernet.h"
-#include "nm-hso-gsm-device.h"
-#include "nm-gsm-device.h"
-#include "nm-cdma-device.h"
 #include "nm-dbus-manager.h"
 #include "nm-setting-ip4-config.h"
 #include "nm-setting-connection.h"
 #include "NetworkManagerSystem.h"
 #include "nm-named-manager.h"
 #include "nm-vpn-manager.h"
+#include "nm-modem.h"
 
 typedef struct LookupThread LookupThread;
 
@@ -234,8 +232,7 @@ get_best_device (NMManager *manager, NMActRequest **out_req)
 			}
 		}
 
-		/* 'hso' devices never get a gateway from the remote end */
-		if (!can_default && !NM_IS_HSO_GSM_DEVICE (dev))
+		if (!can_default && !NM_IS_MODEM (dev))
 			continue;
 
 		/* 'never-default' devices can't ever be the default */
@@ -253,84 +250,13 @@ get_best_device (NMManager *manager, NMActRequest **out_req)
 	return best;
 }
 
-static gboolean
-is_localhost_mapping (const char *str)
-{
-	return (!strncmp (str, "127.0.0.1", strlen ("127.0.0.1")) && strstr (str, "localhost"));
-}
-
-static gboolean
-find_token (const char *line, const char *token)
-{
-	const char *start = line, *p = line;
-
-	g_return_val_if_fail (line != NULL, FALSE);
-	g_return_val_if_fail (token != NULL, FALSE);
-	g_return_val_if_fail (strlen (token) > 0, FALSE);
-
-	/* Walk through the line to find the next whitespace character */
-	while (p <= line + strlen (line)) {
-		if (isblank (*p) || (*p == '\0')) {
-			/* Token starts with 'start' and ends with 'end' */
-			if ((p > start) && *start && !strncmp (start, token, (p - start)))
-				return TRUE; /* found */
-
-			/* not found; advance start and continue looking */
-			start = p + 1;
-		}
-		p++;
-	}
-
-	return FALSE;
-}
-
-#if 0
-/* Testcase for find_token; break it out and add it to the testsuite */
-
-typedef struct {
-	const char *line;
-	const char *token;
-	gboolean expected;
-} Foo;
-
-static Foo foo[] = {
-	{ "127.0.0.1\tfoobar\tblah", "blah", TRUE },
-	{ "", "blah", FALSE },
-	{ "1.1.1.1\tbork\tfoo", "blah", FALSE },
-	{ "127.0.0.1 foobar\tblah", "blah", TRUE },
-	{ "127.0.0.1 foobar blah", "blah", TRUE },
-	{ "192.168.1.1 blah borkbork", "blah", TRUE },
-	{ "192.168.1.1 foobar\tblah borkbork", "blah", TRUE },
-	{ "192.168.1.1\tfoobar\tblah\tborkbork", "blah", TRUE },
-	{ "192.168.1.1 \tfoobar \tblah \tborkbork\t ", "blah", TRUE },
-	{ "\t\t\t\t   \t\t\tasdfadf  a\t\t\t\t\t   \t\t\t\t\t ", "blah", FALSE },
-	{ NULL, NULL, FALSE }
-};
-
-int main(int argc, char **argv)
-{
-	Foo *iter = &foo[0];
-
-	while (iter->line) {
-		if (find_token (iter->line, iter->token) != iter->expected) {
-			g_message ("Failed: '%s' <= '%s' (%d)", iter->line, iter->token, iter->expected);
-			return 1;
-		}
-		iter++;
-	}
-
-	g_message ("Success");
-	return 0;
-}
-#endif
-
 #define FALLBACK_HOSTNAME "localhost.localdomain"
 
 static gboolean
 update_etc_hosts (const char *hostname)
 {
 	char *contents = NULL;
-	char **lines = NULL, **line, **host_mapping = NULL;
+	char **lines = NULL, **line;
 	GError *error = NULL;
 	gboolean initial_comments = TRUE;
 	gboolean added = FALSE;
@@ -357,62 +283,25 @@ update_etc_hosts (const char *hostname)
 		return FALSE;
 	}
 
-	/* Two-pass modification of /etc/hosts:
-	 *
-	 * 1) Look for a non-comment, non-localhost line that contains the current
-	 *    hostname.  Mark that line.
-	 *
-	 * 2) For each line in the existing /etc/hosts, add it to the new /etc/hosts
-	 *    unless it starts with 127.0.0.1 and is right after the initial comments
-	 *    (if any) and contains "localhost".
+	/* Replace any 127.0.0.1 entry that is at the beginning of the file or right
+	 * after initial comments.  If there is no 127.0.0.1 entry at the beginning
+	 * or after initial comments, add one there and ignore any other 127.0.0.1
+	 * entries.
 	 */
-
-	/* Find any existing hostname mapping */
 	for (line = lines; lines && *line; line++) {
-		/* Look for any line that (a) contains the current hostname, and
-		 * (b) does not start with '127.0.0.1' and contain 'localhost'.
-		 */
-		if (   strlen (*line)
-		    && (*line[0] != '#')
-		    && find_token (*line, hostname)
-		    && !is_localhost_mapping (*line)) {
-			host_mapping = line;
-			break;
-		}
-	}
-
-	/* Construct the new hosts file; replace any 127.0.0.1 entry that is at the
-	 * beginning of the file or right after initial comments and contains
-	 * the string 'localhost'.  If there is no 127.0.0.1 entry at the beginning
-	 * or after initial comments that contains 'localhost', add one there
-	 * and ignore any other 127.0.0.1 entries that contain 'localhost'.
-	 */
-	for (line = lines, initial_comments = TRUE; lines && *line; line++) {
 		gboolean add_line = TRUE;
 
 		/* This is the first line after the initial comments */
-		if (strlen (*line) && initial_comments && (*line[0] != '#')) {
+		if (initial_comments && (*line[0] != '#')) {
 			initial_comments = FALSE;
-
-			/* If some other line contained the hostname, make a simple
-			 * localhost mapping and assume the user knows what they are doing
-			 * with their manual hostname entry.  Otherwise if the hostname
-			 * wasn't found somewhere else, add it to the localhost mapping line
-			 * to make sure it's mapped to something.
-			 */
-			if (host_mapping)
-				g_string_append (new_contents, "127.0.0.1");
-			else
-				g_string_append_printf (new_contents, "127.0.0.1\t%s", hostname);
-
+			g_string_append_printf (new_contents, "127.0.0.1\t%s", hostname);
 			if (strcmp (hostname, FALLBACK_HOSTNAME))
 				g_string_append_printf (new_contents, "\t" FALLBACK_HOSTNAME);
-
 			g_string_append (new_contents, "\tlocalhost\n");
 			added = TRUE;
 
 			/* Don't add the entry if it's supposed to be the actual localhost reverse mapping */
-			if (is_localhost_mapping (*line))
+			if (!strncmp (*line, "127.0.0.1", strlen ("127.0.0.1")) && strstr (*line, "localhost"))
 				add_line = FALSE;
 		}
 
@@ -759,7 +648,7 @@ auto_activate_device (gpointer user_data)
 		GError *error = NULL;
 		const char *device_path;
 
-		device_path = nm_device_get_udi (data->device);
+		device_path = nm_device_get_path (data->device);
 		if (!nm_manager_activate_connection (policy->manager,
 		                                     best_connection,
 		                                     specific_object,
@@ -821,7 +710,25 @@ hostname_changed (NMManager *manager, GParamSpec *pspec, gpointer user_data)
 }
 
 static void
-schedule_activate_check (NMPolicy *policy, NMDevice *device)
+sleeping_changed (NMManager *manager, GParamSpec *pspec, gpointer user_data)
+{
+	gboolean sleeping = FALSE;
+	GSList *connections, *iter;
+
+	g_object_get (G_OBJECT (manager), NM_MANAGER_SLEEPING, &sleeping, NULL);
+
+	/* Clear the invalid flag on all connections so they'll get retried on wakeup */
+	if (sleeping) {
+		connections = nm_manager_get_connections (manager, NM_CONNECTION_SCOPE_SYSTEM);
+		connections = g_slist_concat (connections, nm_manager_get_connections (manager, NM_CONNECTION_SCOPE_USER));
+		for (iter = connections; iter; iter = g_slist_next (iter))
+			g_object_set_data (G_OBJECT (iter->data), INVALID_TAG, NULL);
+		g_slist_free (connections);
+	}
+}
+
+static void
+schedule_activate_check (NMPolicy *policy, NMDevice *device, guint delay_seconds)
 {
 	ActivateData *data;
 	GSList *iter;
@@ -834,7 +741,7 @@ schedule_activate_check (NMPolicy *policy, NMDevice *device)
 	if (state < NM_DEVICE_STATE_DISCONNECTED)
 		return;
 
-	if (!nm_device_can_activate (device) || !nm_device_autoconnect_allowed (device))
+	if (!nm_device_autoconnect_allowed (device))
 		return;
 
 	for (iter = policy->pending_activation_checks; iter; iter = g_slist_next (iter)) {
@@ -848,7 +755,7 @@ schedule_activate_check (NMPolicy *policy, NMDevice *device)
 
 	data->policy = policy;
 	data->device = g_object_ref (device);
-	data->id = g_idle_add (auto_activate_device, data);
+	data->id = delay_seconds ? g_timeout_add_seconds (delay_seconds, auto_activate_device, data) : g_idle_add (auto_activate_device, data);
 	policy->pending_activation_checks = g_slist_append (policy->pending_activation_checks, data);
 }
 
@@ -876,13 +783,15 @@ device_state_changed (NMDevice *device,
 
 	switch (new_state) {
 	case NM_DEVICE_STATE_FAILED:
-		/* Mark the connection invalid so it doesn't get automatically chosen */
-		if (connection) {
+		/* Mark the connection invalid if it failed during activation so that
+		 * it doesn't get automatically chosen over and over and over again.
+		 */
+		if (connection && IS_ACTIVATING_STATE (old_state)) {
 			g_object_set_data (G_OBJECT (connection), INVALID_TAG, GUINT_TO_POINTER (TRUE));
 			nm_info ("Marking connection '%s' invalid.", get_connection_id (connection));
 			nm_connection_clear_secrets (connection);
 		}
-		schedule_activate_check (policy, device);
+		schedule_activate_check (policy, device, 3);
 		break;
 	case NM_DEVICE_STATE_ACTIVATED:
 		/* Clear the invalid tag on the connection */
@@ -895,7 +804,7 @@ device_state_changed (NMDevice *device,
 	case NM_DEVICE_STATE_UNAVAILABLE:
 	case NM_DEVICE_STATE_DISCONNECTED:
 		update_routing_and_dns (policy, FALSE);
-		schedule_activate_check (policy, device);
+		schedule_activate_check (policy, device, 0);
 		break;
 	default:
 		break;
@@ -913,7 +822,7 @@ device_ip4_config_changed (NMDevice *device,
 static void
 wireless_networks_changed (NMDeviceWifi *device, NMAccessPoint *ap, gpointer user_data)
 {
-	schedule_activate_check ((NMPolicy *) user_data, NM_DEVICE (device));
+	schedule_activate_check ((NMPolicy *) user_data, NM_DEVICE (device), 0);
 }
 
 typedef struct {
@@ -1009,7 +918,7 @@ schedule_activate_all (NMPolicy *policy)
 
 	devices = nm_manager_get_devices (policy->manager);
 	for (iter = devices; iter; iter = g_slist_next (iter))
-		schedule_activate_check (policy, NM_DEVICE (iter->data));
+		schedule_activate_check (policy, NM_DEVICE (iter->data), 0);
 }
 
 static void
@@ -1099,10 +1008,14 @@ nm_policy_new (NMManager *manager, NMVPNManager *vpn_manager)
 	                       G_CALLBACK (global_state_changed), policy);
 	policy->signal_ids = g_slist_append (policy->signal_ids, (gpointer) id);
 
-	id = g_signal_connect (manager, "notify::hostname",
+	id = g_signal_connect (manager, "notify::" NM_MANAGER_HOSTNAME,
 	                       G_CALLBACK (hostname_changed), policy);
 	policy->signal_ids = g_slist_append (policy->signal_ids, (gpointer) id);
 
+	id = g_signal_connect (manager, "notify::" NM_MANAGER_SLEEPING,
+	                       G_CALLBACK (sleeping_changed), policy);
+	policy->signal_ids = g_slist_append (policy->signal_ids, (gpointer) id);
+
 	id = g_signal_connect (manager, "device-added",
 	                       G_CALLBACK (device_added), policy);
 	policy->signal_ids = g_slist_append (policy->signal_ids, (gpointer) id);
