@@ -23,6 +23,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
+#include <sys/wait.h>
 #include <ctype.h>
 #include <sys/inotify.h>
 #include <errno.h>
@@ -182,6 +183,284 @@ make_connection_setting (const char *file,
 	return NM_SETTING (s_con);
 }
 
+static gboolean
+read_mac_address (shvarFile *ifcfg, GByteArray **array, GError **error)
+{
+	char *value = NULL;
+	struct ether_addr *mac;
+
+	g_return_val_if_fail (ifcfg != NULL, FALSE);
+	g_return_val_if_fail (array != NULL, FALSE);
+	g_return_val_if_fail (*array == NULL, FALSE);
+	g_return_val_if_fail (error != NULL, FALSE);
+	g_return_val_if_fail (*error == NULL, FALSE);
+
+	value = svGetValue (ifcfg, "HWADDR", FALSE);
+	if (!value || !strlen (value)) {
+		g_free (value);
+		return TRUE;
+	}
+
+	mac = ether_aton (value);
+	if (!mac) {
+		g_free (value);
+		g_set_error (error, ifcfg_plugin_error_quark (), 0,
+		             "The MAC address '%s' was invalid.", value);
+		return FALSE;
+	}
+
+	g_free (value);
+	*array = g_byte_array_sized_new (ETH_ALEN);
+	g_byte_array_append (*array, (guint8 *) mac->ether_addr_octet, ETH_ALEN);
+	return TRUE;
+}
+
+static void
+iscsiadm_child_setup (gpointer user_data G_GNUC_UNUSED)
+{
+	/* We are in the child process here; set a different process group to
+	 * ensure signal isolation between child and parent.
+	 */
+	pid_t pid = getpid ();
+	setpgid (pid, pid);
+}
+
+static char *
+match_iscsiadm_tag (const char *line, const char *tag, gboolean *skip)
+{
+	char *p;
+
+	if (g_ascii_strncasecmp (line, tag, strlen (tag)))
+		return NULL;
+
+	p = strchr (line, '=');
+	if (!p) {
+		g_warning ("%s: malformed iscsiadm record: no = in '%s'.",
+		           __func__, line);
+		*skip = TRUE;
+		return NULL;
+	}
+
+	p++; /* advance past = */
+	return g_strstrip (p);
+}
+
+#define ISCSI_HWADDR_TAG    "iface.hwaddress"
+#define ISCSI_BOOTPROTO_TAG "iface.bootproto"
+#define ISCSI_IPADDR_TAG    "iface.ipaddress"
+#define ISCSI_SUBNET_TAG    "iface.subnet_mask"
+#define ISCSI_GATEWAY_TAG   "iface.gateway"
+#define ISCSI_DNS1_TAG      "iface.primary_dns"
+#define ISCSI_DNS2_TAG      "iface.secondary_dns"
+
+static gboolean
+fill_ip4_setting_from_ibft (shvarFile *ifcfg,
+                            NMSettingIP4Config *s_ip4,
+                            const char *iscsiadm_path,
+                            GError **error)
+{
+	const char *argv[4] = { iscsiadm_path, "-m", "fw", NULL };
+	const char *envp[1] = { NULL };
+	gboolean success = FALSE, in_record = FALSE, hwaddr_matched = FALSE, skip = FALSE;
+	char *out = NULL, *err = NULL;
+	gint status = 0;
+	GByteArray *ifcfg_mac = NULL;
+	char **lines = NULL, **iter;
+	const char *method = NULL;
+	struct in_addr ipaddr;
+	struct in_addr gateway;
+	struct in_addr dns1;
+	struct in_addr dns2;
+	guint32 prefix = 0;
+
+	g_return_val_if_fail (s_ip4 != NULL, FALSE);
+	g_return_val_if_fail (iscsiadm_path != NULL, FALSE);
+
+	if (!g_spawn_sync ("/", (char **) argv, (char **) envp, 0,
+	                   iscsiadm_child_setup, NULL, &out, &err, &status, error))
+		return FALSE;
+
+	if (!WIFEXITED (status)) {
+		g_set_error (error, ifcfg_plugin_error_quark (), 0,
+		             "%s exited abnormally.", iscsiadm_path);
+		goto done;
+	}
+
+	if (WEXITSTATUS (status) != 0) {
+		g_set_error (error, ifcfg_plugin_error_quark (), 0,
+		             "%s exited with error %d.  Message: '%s'",
+		             iscsiadm_path, WEXITSTATUS (status), err ? err : "(none)");
+		goto done;
+	}
+
+	if (!read_mac_address (ifcfg, &ifcfg_mac, error))
+		goto done;
+	/* Ensure we got a MAC */
+	if (!ifcfg_mac) {
+		g_set_error (error, ifcfg_plugin_error_quark (), 0,
+		             "Missing device MAC address (no HWADDR tag present).");
+		goto done;
+	}
+
+	memset (&ipaddr, 0, sizeof (ipaddr));
+	memset (&gateway, 0, sizeof (gateway));
+	memset (&dns1, 0, sizeof (dns1));
+	memset (&dns2, 0, sizeof (dns2));
+
+	/* Success, lets parse the output */
+	lines = g_strsplit_set (out, "\n\r", -1);
+	for (iter = lines; iter && *iter; iter++) {
+		char *p;
+
+		if (!g_ascii_strcasecmp (*iter, "# BEGIN RECORD")) {
+			if (in_record) {
+				g_warning ("%s: malformed iscsiadm record: already parsing record.", __func__);
+				skip = TRUE;
+			}
+		} else if (!g_ascii_strcasecmp (*iter, "# END RECORD")) {
+			if (!skip && hwaddr_matched) {
+				/* Record is good; fill IP4 config with its info */
+				if (!method) {
+					g_warning ("%s: malformed iscsiadm record: missing BOOTPROTO.", __func__);
+					return FALSE;
+				}
+
+				g_object_set (s_ip4, NM_SETTING_IP4_CONFIG_METHOD, method, NULL);
+
+				if (!strcmp (method, NM_SETTING_IP4_CONFIG_METHOD_MANUAL)) {
+					NMIP4Address *addr;
+
+				    if (!ipaddr.s_addr || !prefix) {
+						g_warning ("%s: malformed iscsiadm record: BOOTPROTO=static "
+						           "but missing IP address or prefix.", __func__);
+						return FALSE;
+					}
+
+					addr = nm_ip4_address_new ();
+					nm_ip4_address_set_address (addr, ipaddr.s_addr);
+					nm_ip4_address_set_prefix (addr, prefix);
+					nm_ip4_address_set_gateway (addr, gateway.s_addr);
+					nm_setting_ip4_config_add_address (s_ip4, addr);
+					nm_ip4_address_unref (addr);
+
+					if (dns1.s_addr)
+						nm_setting_ip4_config_add_dns (s_ip4, dns1.s_addr);
+					if (dns2.s_addr)
+						nm_setting_ip4_config_add_dns (s_ip4, dns2.s_addr);
+
+					// FIXME: DNS search domains?
+				}
+				return TRUE;
+			}
+			skip = FALSE;
+			hwaddr_matched = FALSE;
+			memset (&ipaddr, 0, sizeof (ipaddr));
+			memset (&gateway, 0, sizeof (gateway));
+			memset (&dns1, 0, sizeof (dns1));
+			memset (&dns2, 0, sizeof (dns2));
+			prefix = 0;
+			method = NULL;
+		}
+
+		if (skip)
+			continue;
+
+		/* HWADDR */
+		if (!skip && (p = match_iscsiadm_tag (*iter, ISCSI_HWADDR_TAG, &skip))) {
+			struct ether_addr *ibft_mac;
+
+			ibft_mac = ether_aton (p);
+			if (!ibft_mac) {
+				g_warning ("%s: malformed iscsiadm record: invalid hwaddress.", __func__);
+				skip = TRUE;
+				continue;
+			}
+
+			if (memcmp (ifcfg_mac->data, (guint8 *) ibft_mac->ether_addr_octet, ETH_ALEN)) {
+				/* This record isn't for the current device, ignore it */
+				skip = TRUE;
+				continue;
+			}
+
+			/* Success, this record is for this device */
+			hwaddr_matched = TRUE;
+		}
+
+		/* BOOTPROTO */
+		if (!skip && (p = match_iscsiadm_tag (*iter, ISCSI_BOOTPROTO_TAG, &skip))) {
+			if (!g_ascii_strcasecmp (p, "dhcp"))
+				method = NM_SETTING_IP4_CONFIG_METHOD_AUTO;
+			else if (!g_ascii_strcasecmp (p, "static"))
+				method = NM_SETTING_IP4_CONFIG_METHOD_MANUAL;
+			else {
+				g_warning ("%s: malformed iscsiadm record: unknown BOOTPROTO '%s'.",
+				           __func__, p);
+				skip = TRUE;
+				continue;
+			}
+		}
+
+		if (!skip && (p = match_iscsiadm_tag (*iter, ISCSI_IPADDR_TAG, &skip))) {
+			if (inet_pton (AF_INET, p, &ipaddr) < 1) {
+				g_warning ("%s: malformed iscsiadm record: invalid IP address '%s'.",
+				           __func__, p);
+				skip = TRUE;
+				continue;
+			}
+		}
+
+		if (!skip && (p = match_iscsiadm_tag (*iter, ISCSI_SUBNET_TAG, &skip))) {
+			struct in_addr mask;
+
+			if (inet_pton (AF_INET, p, &mask) < 1) {
+				g_warning ("%s: malformed iscsiadm record: invalid subnet mask '%s'.",
+				           __func__, p);
+				skip = TRUE;
+				continue;
+			}
+
+			prefix = nm_utils_ip4_netmask_to_prefix (mask.s_addr);
+		}
+
+		if (!skip && (p = match_iscsiadm_tag (*iter, ISCSI_GATEWAY_TAG, &skip))) {
+			if (inet_pton (AF_INET, p, &gateway) < 1) {
+				g_warning ("%s: malformed iscsiadm record: invalid IP gateway '%s'.",
+				           __func__, p);
+				skip = TRUE;
+				continue;
+			}
+		}
+
+		if (!skip && (p = match_iscsiadm_tag (*iter, ISCSI_DNS1_TAG, &skip))) {
+			if (inet_pton (AF_INET, p, &dns1) < 1) {
+				g_warning ("%s: malformed iscsiadm record: invalid DNS1 address '%s'.",
+				           __func__, p);
+				skip = TRUE;
+				continue;
+			}
+		}
+
+		if (!skip && (p = match_iscsiadm_tag (*iter, ISCSI_DNS2_TAG, &skip))) {
+			if (inet_pton (AF_INET, p, &dns2) < 1) {
+				g_warning ("%s: malformed iscsiadm record: invalid DNS2 address '%s'.",
+				           __func__, p);
+				skip = TRUE;
+				continue;
+			}
+		}
+	}
+
+	success = TRUE;
+
+done:
+	if (ifcfg_mac)
+		g_byte_array_free (ifcfg_mac, TRUE);
+	g_strfreev (lines);
+	g_free (out);
+	g_free (err);
+	return success;
+}
+
 static gboolean
 read_ip4_address (shvarFile *ifcfg,
                   const char *tag,
@@ -299,19 +578,6 @@ read_full_ip4_address (shvarFile *ifcfg,
 		nm_ip4_address_set_prefix (addr, nm_utils_ip4_netmask_to_prefix (tmp));
 	}
 
-	/* Try to autodetermine the prefix for the address' class */
-	if (!nm_ip4_address_get_prefix (addr)) {
-		guint32 prefix = 0;
-
-		prefix = nm_utils_ip4_get_default_prefix (nm_ip4_address_get_address (addr));
-		nm_ip4_address_set_prefix (addr, prefix);
-
-		value = svGetValue (ifcfg, ip_tag, FALSE);
-		PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: missing %s, assuming %s/%u",
-		             prefix_tag, value, prefix);
-		g_free (value);
-	}
-
 	/* Validate the prefix */
 	if (  !nm_ip4_address_get_prefix (addr)
 	    || nm_ip4_address_get_prefix (addr) > 32) {
@@ -587,7 +853,10 @@ error:
 
 
 static NMSetting *
-make_ip4_setting (shvarFile *ifcfg, const char *network_file, GError **error)
+make_ip4_setting (shvarFile *ifcfg,
+                  const char *network_file,
+                  const char *iscsiadm_path,
+                  GError **error)
 {
 	NMSettingIP4Config *s_ip4 = NULL;
 	char *value = NULL;
@@ -636,7 +905,15 @@ make_ip4_setting (shvarFile *ifcfg, const char *network_file, GError **error)
 	if (value) {
 		if (!g_ascii_strcasecmp (value, "bootp") || !g_ascii_strcasecmp (value, "dhcp"))
 			method = NM_SETTING_IP4_CONFIG_METHOD_AUTO;
-		else if (!g_ascii_strcasecmp (value, "autoip")) {
+		else if (!g_ascii_strcasecmp (value, "ibft")) {
+			/* iSCSI Boot Firmware Table: need to read values from the iSCSI 
+			 * firmware for this device and create the IP4 setting using those.
+			 */
+			if (fill_ip4_setting_from_ibft (ifcfg, s_ip4, iscsiadm_path, error))
+				return NM_SETTING (s_ip4);
+			g_object_unref (s_ip4);
+			return NULL;
+		} else if (!g_ascii_strcasecmp (value, "autoip")) {
 			g_free (value);
 			g_object_set (s_ip4,
 			              NM_SETTING_IP4_CONFIG_METHOD, NM_SETTING_IP4_CONFIG_METHOD_LINK_LOCAL,
@@ -814,38 +1091,6 @@ error:
 	return NULL;
 }
 
-static gboolean
-read_mac_address (shvarFile *ifcfg, GByteArray **array, GError **error)
-{
-	char *value = NULL;
-	struct ether_addr *mac;
-
-	g_return_val_if_fail (ifcfg != NULL, FALSE);
-	g_return_val_if_fail (array != NULL, FALSE);
-	g_return_val_if_fail (*array == NULL, FALSE);
-	g_return_val_if_fail (error != NULL, FALSE);
-	g_return_val_if_fail (*error == NULL, FALSE);
-
-	value = svGetValue (ifcfg, "HWADDR", FALSE);
-	if (!value || !strlen (value)) {
-		g_free (value);
-		return TRUE;
-	}
-
-	mac = ether_aton (value);
-	if (!mac) {
-		g_free (value);
-		g_set_error (error, ifcfg_plugin_error_quark (), 0,
-		             "The MAC address '%s' was invalid.", value);
-		return FALSE;
-	}
-
-	g_free (value);
-	*array = g_byte_array_sized_new (ETH_ALEN);
-	g_byte_array_append (*array, (guint8 *) mac->ether_addr_octet, ETH_ALEN);
-	return TRUE;
-}
-
 static gboolean
 add_one_wep_key (shvarFile *ifcfg,
                  const char *shvar_key,
@@ -1114,6 +1359,7 @@ parse_wpa_psk (shvarFile *ifcfg,
 {
 	shvarFile *keys_ifcfg;
 	char *psk = NULL, *p, *hashed = NULL;
+	gboolean quoted = FALSE;
 
 	/* Passphrase must be between 10 and 66 characters in length becuase WPA
 	 * hex keys are exactly 64 characters (no quoting), and WPA passphrases
@@ -1139,21 +1385,11 @@ parse_wpa_psk (shvarFile *ifcfg,
 	}
 
 	p = psk;
-	if (p[0] == '"' && psk[strlen (psk) - 1] == '"') {
-		/* Get rid of the quotes */
-		p++;
-		p[strlen (p) - 1] = '\0';
 
-		/* Length check */
-		if (strlen (p) < 8 || strlen (p) > 63) {
-			g_set_error (error, ifcfg_plugin_error_quark (), 0,
-			             "Invalid WPA_PSK (passphrases must be between "
-			             "8 and 63 characters long (inclusive))");
-			goto out;
-		}
+	if (p[0] == '"' && psk[strlen (psk) - 1] == '"')
+		quoted = TRUE;
 
-		hashed = g_strdup (p);
-	} else if (strlen (psk) == 64) {
+	if (!quoted && (strlen (psk) == 64)) {
 		/* Verify the hex PSK; 64 digits */
 		while (*p) {
 			if (!isxdigit (*p++)) {
@@ -1164,6 +1400,30 @@ parse_wpa_psk (shvarFile *ifcfg,
 		}
 		hashed = g_strdup (psk);
 	} else {
+		/* Prior to 4f6eef9e77265484555663cf666cde4fa8323469 and
+		 * 28e2e446868b94b92edc4a82aa0bf1e3eda8ec54 the writer may not have
+		 * properly quoted passphrases, so just handle anything that's unquoted
+		 * and between 8 and 63 characters as a passphrase.
+		 */
+
+		if (quoted) {
+			/* Get rid of the quotes */
+			p++;
+			p[strlen (p) - 1] = '\0';
+		}
+
+		/* Length check */
+		if (strlen (p) < 8 || strlen (p) > 63) {
+			g_set_error (error, ifcfg_plugin_error_quark (), 0,
+			             "Invalid WPA_PSK (passphrases must be between "
+			             "8 and 63 characters long (inclusive))");
+			goto out;
+		}
+
+		hashed = g_strdup (p);
+	}
+
+	if (!hashed) {
 		g_set_error (error, ifcfg_plugin_error_quark (), 0,
 		             "Invalid WPA_PSK (doesn't look like a passphrase or hex key)");
 		goto out;
@@ -1259,21 +1519,6 @@ get_cert_file (const char *ifcfg_path, const char *cert_path)
 	return ret;
 }
 
-static void
-set_file_path (NMSetting8021x *s_8021x,
-               const char *path_tag,
-               const char *hash_tag,
-               const char *path,
-               const char *setting_key)
-{
-	GByteArray *data = NULL;
-
-	g_object_set_data_full (G_OBJECT (s_8021x), path_tag, g_strdup (path), g_free);
-	g_object_get (G_OBJECT (s_8021x), setting_key, &data, NULL);
-	if (data)
-		g_object_set_data_full (G_OBJECT (s_8021x), hash_tag, utils_hash_byte_array (data), g_free);
-}
-
 static gboolean
 eap_tls_reader (const char *eap_method,
                 shvarFile *ifcfg,
@@ -1289,7 +1534,7 @@ eap_tls_reader (const char *eap_method,
 	char *privkey = NULL;
 	char *privkey_password = NULL;
 	gboolean success = FALSE;
-	NMSetting8021xCKType privkey_type = NM_SETTING_802_1X_CK_TYPE_UNKNOWN;
+	NMSetting8021xCKFormat privkey_format = NM_SETTING_802_1X_CK_FORMAT_UNKNOWN;
 
 	value = svGetValue (ifcfg, "IEEE_8021X_IDENTITY", FALSE);
 	if (!value) {
@@ -1307,21 +1552,19 @@ eap_tls_reader (const char *eap_method,
 	if (ca_cert) {
 		real_path = get_cert_file (ifcfg->fileName, ca_cert);
 		if (phase2) {
-			if (!nm_setting_802_1x_set_phase2_ca_cert_from_file (s_8021x, real_path, NULL, error))
+			if (!nm_setting_802_1x_set_phase2_ca_cert (s_8021x,
+			                                           real_path,
+			                                           NM_SETTING_802_1X_CK_SCHEME_PATH,
+			                                           NULL,
+			                                           error))
 				goto done;
-			set_file_path (s_8021x,
-			               TAG_PHASE2_CA_CERT_PATH,
-			               TAG_PHASE2_CA_CERT_HASH,
-			               real_path,
-			               NM_SETTING_802_1X_PHASE2_CA_CERT);
 		} else {
-			if (!nm_setting_802_1x_set_ca_cert_from_file (s_8021x, real_path, NULL, error))
+			if (!nm_setting_802_1x_set_ca_cert (s_8021x,
+			                                    real_path,
+			                                    NM_SETTING_802_1X_CK_SCHEME_PATH,
+			                                    NULL,
+			                                    error))
 				goto done;
-			set_file_path (s_8021x,
-			               TAG_CA_CERT_PATH,
-			               TAG_CA_CERT_HASH,
-			               real_path,
-			               NM_SETTING_802_1X_CA_CERT);
 		}
 	} else {
 		PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: missing %s for EAP"
@@ -1364,56 +1607,30 @@ eap_tls_reader (const char *eap_method,
 	g_free (real_path);
 	real_path = get_cert_file (ifcfg->fileName, privkey);
 	if (phase2) {
-		if (!nm_setting_802_1x_set_phase2_private_key_from_file (s_8021x, real_path, privkey_password, &privkey_type, error))
+		if (!nm_setting_802_1x_set_phase2_private_key (s_8021x,
+		                                               real_path,
+		                                               privkey_password,
+			                                           NM_SETTING_802_1X_CK_SCHEME_PATH,
+		                                               &privkey_format,
+		                                               error))
 			goto done;
-		set_file_path (s_8021x,
-		               TAG_PHASE2_PRIVATE_KEY_PATH,
-		               TAG_PHASE2_PRIVATE_KEY_HASH,
-		               real_path,
-		               NM_SETTING_802_1X_PHASE2_PRIVATE_KEY);
 	} else {
-		if (!nm_setting_802_1x_set_private_key_from_file (s_8021x, real_path, privkey_password, &privkey_type, error))
+		if (!nm_setting_802_1x_set_private_key (s_8021x,
+		                                        real_path,
+		                                        privkey_password,
+			                                    NM_SETTING_802_1X_CK_SCHEME_PATH,
+		                                        &privkey_format,
+		                                        error))
 			goto done;
-		set_file_path (s_8021x,
-		               TAG_PRIVATE_KEY_PATH,
-		               TAG_PRIVATE_KEY_HASH,
-		               real_path,
-		               NM_SETTING_802_1X_PRIVATE_KEY);
 	}
 
-	/* Per NM requirements, if the private key is pkcs12, set the client cert to the
-	 * same data as the private key, since pkcs12 files contain both.
+	/* Only set the client certificate if the private key is not PKCS#12 format,
+	 * as NM (due to supplicant restrictions) requires.  If the key was PKCS#12,
+	 * then nm_setting_802_1x_set_private_key() already set the client certificate
+	 * to the same value as the private key.
 	 */
-	if (privkey_type == NM_SETTING_802_1X_CK_TYPE_PKCS12) {
-		/* Set the private key password if PKCS#12, because PKCS#12 doesn't get
-		 * decrypted when being stored in the Setting.
-		 */
-		if (phase2)
-			g_object_set (s_8021x, NM_SETTING_802_1X_PHASE2_PRIVATE_KEY_PASSWORD, privkey_password, NULL);
-		else
-			g_object_set (s_8021x, NM_SETTING_802_1X_PRIVATE_KEY_PASSWORD, privkey_password, NULL);
-
-		if (phase2) {
-			if (!nm_setting_802_1x_set_phase2_client_cert_from_file (s_8021x, real_path, NULL, error))
-				goto done;
-			set_file_path (s_8021x,
-			               TAG_PHASE2_CLIENT_CERT_PATH,
-			               TAG_PHASE2_CLIENT_CERT_HASH,
-			               real_path,
-			               NM_SETTING_802_1X_PHASE2_CLIENT_CERT);
-		} else {
-			if (!nm_setting_802_1x_set_client_cert_from_file (s_8021x, real_path, NULL, error))
-				goto done;
-			set_file_path (s_8021x,
-			               TAG_CLIENT_CERT_PATH,
-			               TAG_CLIENT_CERT_HASH,
-			               real_path,
-			               NM_SETTING_802_1X_CLIENT_CERT);
-		}
-	} else {
-		/* Otherwise, private key is "traditional" OpenSSL format, so
-		 * client certificate will be a separate file.
-		 */
+	if (   privkey_format == NM_SETTING_802_1X_CK_FORMAT_RAW_KEY
+	    || privkey_format == NM_SETTING_802_1X_CK_FORMAT_X509) {
 		client_cert = svGetValue (ifcfg,
 		                          phase2 ? "IEEE_8021X_INNER_CLIENT_CERT" : "IEEE_8021X_CLIENT_CERT",
 		                          FALSE);
@@ -1428,21 +1645,19 @@ eap_tls_reader (const char *eap_method,
 		g_free (real_path);
 		real_path = get_cert_file (ifcfg->fileName, client_cert);
 		if (phase2) {
-			if (!nm_setting_802_1x_set_phase2_client_cert_from_file (s_8021x, real_path, NULL, error))
+			if (!nm_setting_802_1x_set_phase2_client_cert (s_8021x,
+			                                               real_path,
+			                                               NM_SETTING_802_1X_CK_SCHEME_PATH,
+			                                               NULL,
+			                                               error))
 				goto done;
-			set_file_path (s_8021x,
-			               TAG_PHASE2_CLIENT_CERT_PATH,
-			               TAG_PHASE2_CLIENT_CERT_HASH,
-			               real_path,
-			               NM_SETTING_802_1X_PHASE2_CLIENT_CERT);
 		} else {
-			if (!nm_setting_802_1x_set_client_cert_from_file (s_8021x, real_path, NULL, error))
+			if (!nm_setting_802_1x_set_client_cert (s_8021x,
+			                                        real_path,
+			                                        NM_SETTING_802_1X_CK_SCHEME_PATH,
+			                                        NULL,
+			                                        error))
 				goto done;
-			set_file_path (s_8021x,
-			               TAG_CLIENT_CERT_PATH,
-			               TAG_CLIENT_CERT_HASH,
-			               real_path,
-			               NM_SETTING_802_1X_CLIENT_CERT);
 		}
 	}
 
@@ -1476,13 +1691,12 @@ eap_peap_reader (const char *eap_method,
 	ca_cert = svGetValue (ifcfg, "IEEE_8021X_CA_CERT", FALSE);
 	if (ca_cert) {
 		real_cert_path = get_cert_file (ifcfg->fileName, ca_cert);
-		if (!nm_setting_802_1x_set_ca_cert_from_file (s_8021x, real_cert_path, NULL, error))
+		if (!nm_setting_802_1x_set_ca_cert (s_8021x,
+		                                    real_cert_path,
+		                                    NM_SETTING_802_1X_CK_SCHEME_PATH,
+		                                    NULL,
+		                                    error))
 			goto done;
-		set_file_path (s_8021x,
-		               TAG_CA_CERT_PATH,
-		               TAG_CA_CERT_HASH,
-		               real_cert_path,
-		               NM_SETTING_802_1X_CA_CERT);
 	} else {
 		PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: missing "
 		             "IEEE_8021X_CA_CERT for EAP method '%s'; this is"
@@ -1578,13 +1792,12 @@ eap_ttls_reader (const char *eap_method,
 	ca_cert = svGetValue (ifcfg, "IEEE_8021X_CA_CERT", FALSE);
 	if (ca_cert) {
 		real_cert_path = get_cert_file (ifcfg->fileName, ca_cert);
-		if (!nm_setting_802_1x_set_ca_cert_from_file (s_8021x, real_cert_path, NULL, error))
+		if (!nm_setting_802_1x_set_ca_cert (s_8021x,
+		                                    real_cert_path,
+		                                    NM_SETTING_802_1X_CK_SCHEME_PATH,
+		                                    NULL,
+		                                    error))
 			goto done;
-		set_file_path (s_8021x,
-		               TAG_CA_CERT_PATH,
-		               TAG_CA_CERT_HASH,
-		               real_cert_path,
-		               NM_SETTING_802_1X_CA_CERT);
 	} else {
 		PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    warning: missing "
 		             "IEEE_8021X_CA_CERT for EAP method '%s'; this is"
@@ -1683,7 +1896,7 @@ fill_8021x (shvarFile *ifcfg,
 		char *lower = NULL;
 
 		lower = g_ascii_strdown (*iter, -1);
-		while (*eap->method && !found) {
+		while (eap->method && !found) {
 			if (strcmp (eap->method, lower))
 				goto next;
 
@@ -1919,7 +2132,8 @@ make_wireless_security_setting (shvarFile *ifcfg,
 
 static NMSetting *
 make_wireless_setting (shvarFile *ifcfg,
-                       gboolean unmanaged,
+                       gboolean nm_controlled,
+                       char **unmanaged,
                        GError **error)
 {
 	NMSettingWireless *s_wireless;
@@ -1931,8 +2145,16 @@ make_wireless_setting (shvarFile *ifcfg,
 	if (read_mac_address (ifcfg, &array, error)) {
 		if (array) {
 			g_object_set (s_wireless, NM_SETTING_WIRELESS_MAC_ADDRESS, array, NULL);
+
+			/* A connection can only be unmanaged if we know the MAC address */
+			if (!nm_controlled) {
+				*unmanaged = g_strdup_printf ("mac:%02x:%02x:%02x:%02x:%02x:%02x",
+				                              array->data[0], array->data[1], array->data[2],
+				                              array->data[3], array->data[4], array->data[5]);
+			}
+
 			g_byte_array_free (array, TRUE);
-		} else if (unmanaged) {
+		} else if (!nm_controlled) {
 			/* If NM_CONTROLLED=no but there wasn't a MAC address, notify
 			 * the user that the device cannot be unmanaged.
 			 */
@@ -2001,13 +2223,13 @@ make_wireless_setting (shvarFile *ifcfg,
 		g_free (value);
 	} else {
 		/* Only fail on lack of SSID if device is managed */
-		if (!unmanaged) {
+		if (nm_controlled) {
 			g_set_error (error, ifcfg_plugin_error_quark (), 0, "Missing SSID");
 			goto error;
 		}
 	}
 
-	if (unmanaged)
+	if (!nm_controlled)
 		goto done;
 
 	value = svGetValue (ifcfg, "MODE", FALSE);
@@ -2098,7 +2320,8 @@ error:
 static NMConnection *
 wireless_connection_from_ifcfg (const char *file,
                                 shvarFile *ifcfg,
-                                gboolean unmanaged,
+                                gboolean nm_controlled,
+                                char **unmanaged,
                                 GError **error)
 {
 	NMConnection *connection = NULL;
@@ -2124,7 +2347,7 @@ wireless_connection_from_ifcfg (const char *file,
 	}
 
 	/* Wireless */
-	wireless_setting = make_wireless_setting (ifcfg, unmanaged, error);
+	wireless_setting = make_wireless_setting (ifcfg, nm_controlled, unmanaged, error);
 	if (!wireless_setting) {
 		g_object_unref (connection);
 		return NULL;
@@ -2137,7 +2360,7 @@ wireless_connection_from_ifcfg (const char *file,
 	else
 		printable_ssid = g_strdup_printf ("unmanaged");
 
-	if (!unmanaged) {
+	if (nm_controlled) {
 		mode = nm_setting_wireless_get_mode (NM_SETTING_WIRELESS (wireless_setting));
 		if (mode && !strcmp (mode, "adhoc"))
 			adhoc = TRUE;
@@ -2172,7 +2395,7 @@ wireless_connection_from_ifcfg (const char *file,
 	nm_connection_add_setting (connection, con_setting);
 
 	/* Don't verify if unmanaged since we may not have an SSID or whatever */
-	if (!unmanaged) {
+	if (nm_controlled) {
 		if (!nm_connection_verify (connection, error)) {
 			g_object_unref (connection);
 			return NULL;
@@ -2185,7 +2408,8 @@ wireless_connection_from_ifcfg (const char *file,
 static NMSetting *
 make_wired_setting (shvarFile *ifcfg,
                     const char *file,
-                    gboolean unmanaged,
+                    gboolean nm_controlled,
+                    char **unmanaged,
                     NMSetting8021x **s_8021x,
                     GError **error)
 {
@@ -2211,8 +2435,16 @@ make_wired_setting (shvarFile *ifcfg,
 	if (read_mac_address (ifcfg, &mac, error)) {
 		if (mac) {
 			g_object_set (s_wired, NM_SETTING_WIRED_MAC_ADDRESS, mac, NULL);
+
+			/* A connection can only be unmanaged if we know the MAC address */
+			if (!nm_controlled) {
+				*unmanaged = g_strdup_printf ("mac:%02x:%02x:%02x:%02x:%02x:%02x",
+				                              mac->data[0], mac->data[1], mac->data[2],
+				                              mac->data[3], mac->data[4], mac->data[5]);
+			}
+
 			g_byte_array_free (mac, TRUE);
-		} else if (!unmanaged) {
+		} else if (!nm_controlled) {
 			/* If NM_CONTROLLED=no but there wasn't a MAC address, notify
 			 * the user that the device cannot be unmanaged.
 			 */
@@ -2248,7 +2480,8 @@ error:
 static NMConnection *
 wired_connection_from_ifcfg (const char *file,
                              shvarFile *ifcfg,
-                             gboolean unmanaged,
+                             gboolean nm_controlled,
+                             char **unmanaged,
                              GError **error)
 {
 	NMConnection *connection = NULL;
@@ -2275,7 +2508,7 @@ wired_connection_from_ifcfg (const char *file,
 	}
 	nm_connection_add_setting (connection, con_setting);
 
-	wired_setting = make_wired_setting (ifcfg, file, unmanaged, &s_8021x, error);
+	wired_setting = make_wired_setting (ifcfg, file, nm_controlled, unmanaged, &s_8021x, error);
 	if (!wired_setting) {
 		g_object_unref (connection);
 		return NULL;
@@ -2336,9 +2569,10 @@ is_wireless_device (const char *iface)
 
 NMConnection *
 connection_from_file (const char *filename,
-                      const char *network_file,
-                      const char *test_type,  /* for unit tests only */
-                      gboolean *ignored,
+                      const char *network_file,  /* for unit tests only */
+                      const char *test_type,     /* for unit tests only */
+                      const char *iscsiadm_path, /* for unit tests only */
+                      char **unmanaged,
                       char **keyfile,
                       char **routefile,
                       GError **error,
@@ -2346,13 +2580,14 @@ connection_from_file (const char *filename,
 {
 	NMConnection *connection = NULL;
 	shvarFile *parsed;
-	char *type;
-	char *nmc = NULL;
+	char *type, *nmc = NULL, *bootproto;
 	NMSetting *s_ip4;
 	const char *ifcfg_name = NULL;
+	gboolean nm_controlled = TRUE;
 
 	g_return_val_if_fail (filename != NULL, NULL);
-	g_return_val_if_fail (ignored != NULL, NULL);
+	g_return_val_if_fail (unmanaged != NULL, NULL);
+	g_return_val_if_fail (*unmanaged == NULL, NULL);
 	g_return_val_if_fail (keyfile != NULL, NULL);
 	g_return_val_if_fail (*keyfile == NULL, NULL);
 	g_return_val_if_fail (routefile != NULL, NULL);
@@ -2362,6 +2597,9 @@ connection_from_file (const char *filename,
 	if (!network_file)
 		network_file = SYSCONFDIR "/sysconfig/network";
 
+	if (!iscsiadm_path)
+		iscsiadm_path = SBINDIR "/iscsiadm";
+
 	ifcfg_name = utils_get_ifcfg_name (filename, TRUE);
 	if (!ifcfg_name) {
 		g_set_error (error, ifcfg_plugin_error_quark (), 0,
@@ -2423,26 +2661,31 @@ connection_from_file (const char *filename,
 		g_free (nmc);
 
 		if (!strcmp (lower, "no") || !strcmp (lower, "n") || !strcmp (lower, "false"))
-			*ignored = TRUE;
+			nm_controlled = FALSE;
 		g_free (lower);
 	}
 
 	if (!strcasecmp (type, TYPE_ETHERNET))
-		connection = wired_connection_from_ifcfg (filename, parsed, *ignored, error);
+		connection = wired_connection_from_ifcfg (filename, parsed, nm_controlled, unmanaged, error);
 	else if (!strcasecmp (type, TYPE_WIRELESS))
-		connection = wireless_connection_from_ifcfg (filename, parsed, *ignored, error);
+		connection = wireless_connection_from_ifcfg (filename, parsed, nm_controlled, unmanaged, error);
 	else {
 		g_set_error (error, ifcfg_plugin_error_quark (), 0,
 		             "Unknown connection type '%s'", type);
 	}
 
+	if (nm_controlled) {
+		g_free (*unmanaged);
+		*unmanaged = NULL;
+	}
+
 	g_free (type);
 
 	/* Don't bother reading the connection fully if it's unmanaged */
-	if (!connection || *ignored)
+	if (!connection || *unmanaged)
 		goto done;
 
-	s_ip4 = make_ip4_setting (parsed, network_file, error);
+	s_ip4 = make_ip4_setting (parsed, network_file, iscsiadm_path, error);
 	if (*error) {
 		g_object_unref (connection);
 		connection = NULL;
@@ -2451,6 +2694,21 @@ connection_from_file (const char *filename,
 		nm_connection_add_setting (connection, s_ip4);
 	}
 
+	/* iSCSI / ibft connections are read-only since their settings are
+	 * stored in NVRAM and can only be changed in BIOS.
+	 */
+	bootproto = svGetValue (parsed, "BOOTPROTO", FALSE);
+	if (   bootproto
+	    && connection
+	    && !g_ascii_strcasecmp (bootproto, "ibft")) {
+		NMSettingConnection *s_con;
+
+		s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
+		g_assert (s_con);
+
+		g_object_set (G_OBJECT (s_con), NM_SETTING_CONNECTION_READ_ONLY, TRUE, NULL);
+	}
+
 	if (!nm_connection_verify (connection, error)) {
 		g_object_unref (connection);
 		connection = NULL;
